import { produceWithPatches } from 'immer';
import { shallow } from 'zustand/shallow';
import { applyDiffWithCreateQueriesFromPatch, applyFlatConfigToState, } from './utils.js';
export var HistoryEventType;
(function (HistoryEventType) {
    HistoryEventType[HistoryEventType["PUSH"] = 0] = "PUSH";
    HistoryEventType[HistoryEventType["REPLACE"] = 1] = "REPLACE";
    HistoryEventType[HistoryEventType["REGISTER"] = 2] = "REGISTER";
})(HistoryEventType || (HistoryEventType = {}));
export const historyManagement = (historyInstance) => (apply) => (set, get, api) => apply((fn, type, ns, options) => {
    return new Promise((resolve, reject) => {
        set((changes, values) => {
            if (changes.length === 0) {
                resolve(null);
                return values;
            }
            if (type !== HistoryEventType.REGISTER) {
                // if namespace is not given, calculate what namespaces are affected
                const affectedNamespaces = ns
                    ? [ns]
                    : changes.reduce((next, change) => {
                        const { path: [, namespace], } = change;
                        if (next.indexOf(namespace) === -1) {
                            return [...next, namespace];
                        }
                        return next;
                    }, []);
                const uniqueQueries = affectedNamespaces.reduce((next, thisNs) => {
                    const { config, query: currentQuery } = get().namespaces[thisNs];
                    return {
                        ...next,
                        [thisNs]: applyDiffWithCreateQueriesFromPatch(config, thisNs, currentQuery, changes, values.namespaces[thisNs].values, values.namespaces[thisNs].initialValues),
                    };
                }, {});
                const method = type === HistoryEventType.PUSH ? 'push' : 'replace';
                const otherQueries = Object.keys(values.namespaces).reduce((next, thisNs) => {
                    if (affectedNamespaces.indexOf(thisNs) !== -1) {
                        return next;
                    }
                    return {
                        ...next,
                        ...values.namespaces[thisNs].query,
                    };
                }, {});
                const reducedQueries = Object.keys(uniqueQueries).reduce((next, thisNs) => ({ ...next, ...uniqueQueries[thisNs] }), {});
                const queryObject = Object.freeze({
                    ...otherQueries,
                    ...reducedQueries,
                });
                historyInstance[method](queryObject, options)
                    .then(resolve)
                    .catch(reject);
                // We save the current state of `query` for all affected namespaces
                return {
                    ...values,
                    namespaces: {
                        ...values.namespaces,
                        ...affectedNamespaces.reduce((next, thisNs) => {
                            return {
                                ...next,
                                [thisNs]: {
                                    ...values.namespaces[thisNs],
                                    query: uniqueQueries[thisNs],
                                },
                            };
                        }, {}),
                    },
                };
            }
            resolve(null);
            return values;
        }, fn, type, ns);
    });
    // H
}, get, api);
/**
 * If a namespace is given, will forward the mutation instead of updating
 * the whole state. Initializes the namespace if it does not exist yet
 */
const namespaceProducer = (fn, ns) => (state) => {
    if (!ns) {
        const result = fn(state.namespaces);
        // if no namespaces is given, we support return values
        if (result) {
            state.namespaces = result;
        }
        return;
    }
    if (state.namespaces[ns]) {
        fn(state.namespaces[ns]);
        return;
    }
    const next = {};
    fn(next);
    state.namespaces[ns] = next;
};
export const immerWithPatches = (config) => (set, get, api) => config((valueMapper, fn, type, ns) => {
    return set((currentState) => {
        const [nextValues, changes] = produceWithPatches(namespaceProducer(fn, ns))(currentState);
        return valueMapper(changes, nextValues);
    }, `action_${HistoryEventType[type]}`);
}, get, api);
const parseSearchString = (search) => typeof search === 'string'
    ? Object.fromEntries(new URLSearchParams(search).entries())
    : Object.fromEntries(search.entries());
export const converter = (historyInstance) => (set, get) => {
    const updateFromQuery = (search) => {
        const nextQueries = parseSearchString(search);
        const namespaces = get().namespaces;
        Object.keys(namespaces).forEach((ns) => {
            // It's possible that the ns got cleared while we are applying the new state.
            // here we explicitly get the reference to the ns, `namespaces` is too weak.
            if (get().namespaces[ns]) {
                const outerState = get().namespaces[ns];
                const thisNextQueries = applyFlatConfigToState(outerState.mappedConfig, nextQueries, ns, {}, outerState.initialValues, false);
                // We might have already the correct state applied that match the query parameters
                if (!shallow(get().namespaces[ns].query, thisNextQueries)) {
                    set((state) => {
                        state.query = applyFlatConfigToState(state.mappedConfig, nextQueries, ns, state.values, state.initialValues);
                    }, HistoryEventType.REGISTER, ns);
                }
            }
        });
    };
    const reset = (ns, event, routerOptions) => set((state) => void (state.values = state.initialValues), event, ns, routerOptions);
    return {
        /** batch pushes the given namespaces */
        batchPushState: (ns, fn, routerOptions) => {
            return set((state) => void fn(...ns.map((thisNs) => (state[thisNs] || {}).values)), HistoryEventType.PUSH, undefined, routerOptions);
        },
        /** batch replaces the given namespaces */
        batchReplaceState: (ns, fn, routerOptions) => {
            return set((state) => void fn(...ns.map((thisNs) => (state[thisNs] || {}).values)), HistoryEventType.REPLACE, undefined, routerOptions);
        },
        /** the initial queries when the script got executed first (usually on page load). */
        initialQueries: () => parseSearchString(historyInstance.initialSearch()),
        /** here we store all data and configurations for the different namespaces */
        namespaces: {},
        /** pushes a new state for a given namespace, (will use history.pushState) */
        pushState: (ns, fn, routerOptions) => set((state) => fn(state.values), HistoryEventType.PUSH, ns, routerOptions),
        /** registers a new namespace and initializes it's configuration */
        register: (config, mappedConfig, ns, initialValues, query, values) => {
            const current = get().namespaces[ns];
            const defaultsEqual = (current === null || current === void 0 ? void 0 : current.initialValues) === initialValues;
            if (current !== undefined) {
                set((state) => {
                    state.subscribers = state.subscribers + 1;
                    if (!defaultsEqual) {
                        state.initialValues = initialValues;
                        state.query = applyFlatConfigToState(state.mappedConfig, parseSearchString(historyInstance.initialSearch()), ns, state.values, initialValues);
                    }
                }, HistoryEventType.REGISTER, ns);
                return {
                    initialValues: get().namespaces[ns].initialValues,
                    unsubscribe: get().namespaces[ns].unsubscribe,
                    values: get().namespaces[ns].values,
                };
            }
            // read initial query state:
            set((state) => {
                state.subscribers = 1;
                state.unsubscribe = () => {
                    ;
                    set((thisState) => {
                        // it's possible that the state namespace has been cleared by the provider
                        if (!thisState[ns]) {
                            return;
                        }
                        thisState[ns].subscribers = thisState[ns].subscribers - 1;
                        if (thisState[ns].subscribers === 0) {
                            delete thisState[ns];
                        }
                    }, HistoryEventType.REGISTER);
                    return !get().namespaces[ns];
                };
                state.mappedConfig = mappedConfig;
                state.config = config;
                state.initialValues = initialValues;
                state.values = values;
                state.query = query;
            }, HistoryEventType.REGISTER, ns);
            return {
                initialValues,
                unsubscribe: get().namespaces[ns].unsubscribe,
                values,
            };
        },
        replaceState: (ns, fn, routerOptions) => set((state) => fn(state.values), HistoryEventType.REPLACE, ns, routerOptions),
        resetPush: (ns, routerOptions) => reset(ns, HistoryEventType.PUSH, routerOptions),
        resetReplace: (ns, routerOptions) => reset(ns, HistoryEventType.REPLACE, routerOptions),
        /** cleans up this instance */
        unregister: () => {
            set(() => {
                // return a new object for namespaces
                return {};
            }, HistoryEventType.REGISTER);
        },
        updateFromQuery,
    };
};
